package rgnavigator.navigation;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import rgnavigator.core.NaviUtil;
import rgnavigator.core.exceptions.WaypointScriptException;
import rgnavigator.core.factories.NavigatorFactory;
import rgnavigator.core.factories.TrackStepFactory;
import rgnavigator.core.interfaces.NaviMapI;
import rgnavigator.core.interfaces.NaviWaypointI;
import rgnavigator.core.interfaces.NavigationContextI;
import rgnavigator.core.interfaces.PosI;
import rgnavigator.core.interfaces.TackNavigatorI;
import rgnavigator.core.interfaces.TrackStepI;
import rgnavigator.core.interfaces.WayConnectorI;
import rgnavigator.core.interfaces.WayNodeI;
import rgnavigator.scripting.NaviWaypointCalculator;

public class TackNavigator implements TackNavigatorI {
	private NavigationContextI navigationContext;
	private NaviMapI naviMap;
	private String profileName;
	private String baseDir;

	// =====================================================================

	public TackNavigator(NaviMapI pNaviMap, String pBaseDir, String pProfileName) {
		profileName = pProfileName;
		baseDir = pBaseDir;
		navigationContext = NavigatorFactory.createNavigationContext();
		setNaviMap(pNaviMap);
	}

	// =====================================================================

	@Override
	public void addStep(PosI pPos) {
		TrackStepI lastStep = null;
		TrackStepI currentStep = null;
		boolean addNewStep = true;
		if (navigationContext.stepCount() > 0) {
			lastStep = navigationContext.getPrevStep(0);
			Collection<WayNodeI> lastPosNeighbours;
			WayConnectorI wayConnector = getNaviMap().getFirstWayConnector(lastStep);
			if (wayConnector != null) {
				lastPosNeighbours = wayConnector.getNeighbourWayNodes();
			} else {
				lastPosNeighbours = lastStep.getNeighbours();
			}
			TrackStepI beforeLastStep = navigationContext.getPrevStep(1);
			// the pos might be off road. Find the best alternative, if it is
			// off road
			pPos = findNextMostLikelyPos(pPos, lastPosNeighbours, lastStep, beforeLastStep);
			// if the best is the last pos => Do not eval this position
			if (!lastStep.equals(pPos)) {
				currentStep = analyseWays(pPos, lastPosNeighbours, lastStep, beforeLastStep);
			} else {
				addNewStep = false;
			}
		}
		if (addNewStep && currentStep == null) {
			WayNodeI wayNode = getNaviMap().getFirstWayNode(pPos);
			currentStep = createTrackStep(lastStep, pPos, wayNode);
		}
		// The last steps and current step can be equal, if we had to guess the
		// position, which was in fact off road
		// ignore equal steps
		if (addNewStep && (lastStep == null || !lastStep.equals(currentStep))) {
			navigationContext.addStep(currentStep);
		}
		navigationContext.incSourceStepIndex();	
	}

	// ==================================================================

	// private Collection<WayNodeI> getAvailableNextSteps(Collection<MapObjectI>
	// pObjects) {
	// Collection<WayNodeI> result = null;
	// if (pObjects != null) {
	// // find the first way node. All the way nodes on this position will
	// // have the same neighbours
	// for (MapObjectI mapObject : pObjects) {
	// if (mapObject instanceof WayNodeI) {
	// WayNodeI wayNode = (WayNodeI) mapObject;
	// result = wayNode.getNeighbours();
	// break;
	// }
	// }
	// }
	// return result;
	// }

	// ==================================================================

	private TrackStepI analyseWays(PosI pCurrentPos, Collection<WayNodeI> pLastPosNeighbours, TrackStepI pLastStep, TrackStepI beforeLastStep) {
		TrackStepI result = null;

		List<TrackStepI> alternatives = new ArrayList<TrackStepI>();
		for (WayNodeI lastPosNeighbourNode : pLastPosNeighbours) {
			if (pCurrentPos.equals(lastPosNeighbourNode)) {
				result = createTrackStep(pLastStep, pCurrentPos, lastPosNeighbourNode);
			} else {

				// Check, if we are still on a way and not off road.
				// If we are off road, we can not check for alternatives
				if (getNaviMap().getFirstWayNode(pCurrentPos) != null) {
					// only get real alternatives. Back the way we came from is
					// no alternative track
					// If the the step before the last one was off route (has no
					// way node) do not evaluate alternatives
					if (beforeLastStep != null && beforeLastStep.getWayNode() != null && !beforeLastStep.equals(lastPosNeighbourNode)) {
						TrackStepI alternativeStep = createTrackStep(pLastStep, lastPosNeighbourNode, lastPosNeighbourNode);
						alternatives.add(alternativeStep);
					}
				} 
			}
		}
		if (pLastStep != null && alternatives.size() > 0) {
			pLastStep.setStepAlternatives(alternatives);
		}
		return result;
	}

	// ==================================================================

	private PosI findNextMostLikelyPos(PosI pCurrentPos, Collection<WayNodeI> pLastPosNeighbours, PosI pLastStep, PosI pBeforeLastStep) {
		PosI result = pCurrentPos;
		if (getNaviMap().getFirstWayNode(pCurrentPos) == null) {
			int bestdist = NaviUtil.calcDistance(pCurrentPos, pLastStep);
			result = pLastStep;
			for (WayNodeI lastPosNeighbourNode : pLastPosNeighbours) {
				if (pBeforeLastStep == null || !pBeforeLastStep.equals(lastPosNeighbourNode)) {
					int dist = NaviUtil.calcDistance(pCurrentPos, lastPosNeighbourNode);
					if (dist < bestdist) {
						result = lastPosNeighbourNode;
						bestdist = dist;
					}
				}
			}
		}
		return result;
	}

	// ==================================================================
	private TrackStepI createTrackStep(TrackStepI pLastStep, PosI pPos, WayNodeI pNode) {
		TrackStepI track = TrackStepFactory.createTrackStep(pPos, pNode);
		double angle = calculateAngleFromContext(pPos);
		track.setAngleChange(angle);
		int len = 0;
		if (pLastStep != null && navigationContext.stepCount() > 0) {
			len = NaviUtil.calcDistance(pLastStep, pPos);
			track.setDistance(pLastStep.getDistance() + len);
		} else {
			track.setDistance(0);
		}
		track.setStepLength(len);
		return track;
	}

	// ==================================================================

	private double calculateAngleFromContext(PosI pPos3) {
		double result = 0;

		if (navigationContext.stepCount() >= 2) {
			result = NaviUtil.calcAngleChange(navigationContext.getPrevStep(1), navigationContext.getPrevStep(0), pPos3);
		}
		return result;
	}

	// ==================================================================

	public NaviMapI getNaviMap() {
		return naviMap;
	}

	// ==================================================================

	public void setNaviMap(NaviMapI naviMap) {
		this.naviMap = naviMap;
	}

	// ==================================================================

	public String getProfileName() {
		return profileName;
	}

	// ==================================================================
	public String getBaseDir() {
		return baseDir;
	}

	// ==================================================================
	@Override
	public List<TrackStepI> getCalculatedTrackSteps() {
		return navigationContext.getSteps();
	}

	// ==================================================================
	@Override
	public List<NaviWaypointI> getNavigationResult() {
		List<NaviWaypointI> result = null;
		NaviWaypointCalculator calculator = new NaviWaypointCalculator();
		String script = getBaseDir() + "/" + getProfileName() + ".scm";
		File scripFile = new File(script);
		if (scripFile.isFile()) {
			calculator.calculateNaviWaypoints(navigationContext, scripFile);
			result = navigationContext.getWaypoints();
		} else {
			throw new WaypointScriptException("Script not found: " + scripFile.getAbsolutePath());
		}
		return result;
	}
}
